import re

from pybtex.plugin import find_plugin
from pybtex.style.formatting import BaseStyle, toplevel
from pybtex.style.template import (
    field, first_of, href, join, optional, optional_field, sentence, tag,
    together, words, node, FieldIsMissing
)
from pybtex.richtext import Text, Symbol

firstlast = find_plugin('pybtex.style.names', 'firstlast')()

def format_pages(text):
    dash_re = re.compile(r'-+')
    pages = Text(Symbol('ndash')).join(text.split(dash_re))
    if re.search('[-‒–—―]', str(text)):
        return Text("pp.", Symbol('nbsp'), pages)
    return Text("p.", Symbol('nbsp'), pages)

pages = field('pages', apply_func=format_pages)
date = words[field('year'), optional[", ", field('month')]]

@node
def apa_names(children, context, role, **kwargs):
    """
    Returns formatted names as an APA compliant reference list citation.
    """
    assert not children

    try:
        persons = context['entry'].persons[role]
    except KeyError:
        raise FieldIsMissing(role, context['entry'])

    style = context['style']

    if len(persons) > 20:
        formatted_names = [style.format_name(
            person, style.abbreviate_names) for person in persons[:20]]
        formatted_names += [richtext.Text("et al.")]
        return join(sep=', ')[formatted_names].format_data(context)
    else:
        formatted_names = [style.format_name(
            person, style.abbreviate_names) for person in persons]
        return join(sep=', ', sep2=' & ', last_sep=', & ')[
            formatted_names].format_data(context)

@node
def editor_names(children, context, with_suffix=True, **kwargs):
    """
    Returns formatted editor names for inbook.
    """
    assert not children

    try:
        editors = context['entry'].persons['editor']
    except KeyError:
        raise FieldIsMissing('editor', context['entry'])

    formatted_names = [
        firstlast.format(editor, True) for editor in editors]

    if with_suffix:
        return words[
            join(sep=', ', sep2=', & ', last_sep=', & ')[formatted_names],
            "(Eds.)" if len(editors) > 1 else "(Ed.)"
        ].format_data(context)

    return join(sep=', ', sep2=', & ', last_sep=', & ')[
        formatted_names
    ].format_data(context)

class APAStyle(BaseStyle):
    name = 'apa7'
    default_name_style = 'lastfirst'
    default_sorting_style = 'author_year_title'
    default_label_style = 'apa7'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.abbreviate_names = True

    def format_names(self, role, as_sentence=True):
        formatted_names = apa_names(role)
        if as_sentence:
            return sentence(capfirst=False)[formatted_names]
        else:
            return formatted_names

    def format_author_or_editor_and_date(self, e):
        if 'author' in e.persons and 'editor' in e.persons:
            return sentence(sep=' ')[
                self.format_names('author'), 
                self.format_date(e),
                self.format_editor(e, as_sentence=False)
            ]
        elif 'author' in e.persons:
            return sentence(sep=' ')[
                self.format_names('author'), 
                self.format_date(e),
            ]
        else:
            return sentence(sep=' ')[
                self.format_editor(e, as_sentence=False),
                self.format_date(e),
            ]

    def format_editor(self, e, as_sentence=True):
        editors = self.format_names('editor', as_sentence=False)
        if 'editor' not in e.persons:
            # when parsing the template, a FieldIsMissing exception
            # will be thrown anyway; no need to do anything now,
            # just return the template that will throw the exception
            return editors
        if len(e.persons['editor']) > 1:
            word = '(Eds.)'
        else:
            word = '(Ed.)'
        result = join(sep=' ')[editors, word]
        if as_sentence:
            return sentence[result]
        else:
            return result

    def format_volume(self, e, for_article=False):
        prefix = "Vol."
        if for_article:
            return join[
                tag('em')[field('volume')],
                optional['(', field('number'), ')'],
            ]
        else:
            return optional[together[prefix, field('volume')]]

    def format_title(self, e, which_field, as_sentence=True):
        formatted_title = field(
            which_field, apply_func=lambda text: text.capitalize()
        )
        if as_sentence:
            return sentence[formatted_title]
        else:
            return formatted_title

    def format_btitle(self, e, which_field, as_sentence=True):
        formatted_title = tag('em')[field(which_field)]
        if as_sentence:
            return sentence[formatted_title]
        else:
            return formatted_title

    def format_web_refs(self, e):
        return sentence(add_period=False)[
            optional[self.format_url(e)],
            optional[self.format_eprint(e)],
            optional[self.format_pubmed(e)],
            optional[self.format_doi(e)],
        ]

    def format_url(self, e):
        return words[
            'URL:',
            href[
                field('url', raw=True),
                field('url', raw=True)
            ]
        ]

    def format_pubmed(self, e):
        return href[
            join[
                'https://www.ncbi.nlm.nih.gov/pubmed/',
                field('pubmed', raw=True)
            ],
            join[
                'PMID:',
                field('pubmed', raw=True)
            ]
        ]

    def format_doi(self, e):
        return href[
            join[
                'https://doi.org/',
                field('doi', raw=True)
            ],
            join[
                'doi:',
                field('doi', raw=True)
            ]
        ]

    def format_eprint(self, e):
        return href[
            join[
                'https://arxiv.org/abs/',
                field('eprint', raw=True)
            ],
            join[
                'arXiv:',
                field('eprint', raw=True)
            ]
        ]

    def format_date(self, e):
        return sentence[
            join["(", first_of[optional[date], "n.d."], ")"]
        ]

    def get_article_template(self, e):
        volume_and_pages = first_of[
            optional[
                join[
                    self.format_volume(e, for_article=True),
                    optional[', ', field('pages')]
                ],
            ],
            pages,
        ]
        if 'author' in e.persons:
            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                self.format_title(e, 'title'),
                sentence[
                    tag('em')[field('journal')],
                    optional[volume_and_pages],
                ],
                sentence[optional_field('note')],
                self.format_web_refs(e),
            ]
        else:
            return toplevel[
                self.format_title(e, 'title'),
                self.format_date(e),
                sentence[
                    tag('em')[field('journal')],
                    optional[volume_and_pages],
                ],
                sentence[optional_field('note')],
                self.format_web_refs(e),
            ]

        return template

    def get_book_template(self, e):
        if 'author' in e.persons or 'editor' in e.persons:
            return toplevel[
                self.format_author_or_editor_and_date(e),
                sentence(sep=' ')[
                    self.format_btitle(e, 'title'),
                    optional[
                        sentence[
                            optional[field('edition'), ' ed.'],
                            self.format_volume(e),
                        ]
                    ]
                ],
                sentence()[
                    field('publisher'),
                ],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                self.format_btitle(e, 'title'),
                self.format_date(e),
                sentence(sep=' ')[
                    optional[
                        sentence[
                            optional[field('edition'), ' ed.'],
                            self.format_volume(e),
                        ]
                    ]
                ],
                sentence()[
                    field('publisher'),
                ],
                sentence[optional_field('note')],
            ]

    def get_booklet_template(self, e):
        if 'author' in e.persons:
            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                self.format_title(e, 'title'),
                sentence[optional_field('address')],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                self.format_title(e, 'title'),
                self.format_date(e),
                sentence[optional_field('address')],
                sentence[optional_field('note')],
            ]

    def get_inbook_template(self, e):
        if 'author' in e.persons:

            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                self.format_title(e, 'title'),
                sentence(sep=' ')[
                    optional["In ", editor_names(), ","],
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional[
                        join[
                            "(",
                            sentence(add_period=False)[
                                optional[field('edition'), ' ed.'],
                                self.format_volume(e),
                                pages,
                            ],
                            ")"
                        ]
                    ]
                ],
                sentence()[
                    field('publisher'),
                ],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                self.format_title(e, 'title'),
                self.format_date(e),
                sentence(sep=' ')[
                    optional["In ", editor_names(), ","],
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional[
                        join[
                            "(",
                            sentence(add_period=False)[
                                optional[field('edition'), ' ed.'],
                                self.format_volume(e),
                                pages,
                            ],
                            ")"
                        ]
                    ]
                ],
                sentence()[
                    field('publisher'),
                ],
                sentence[optional_field('note')],
            ]

    
    def get_incollection_template(self, e):
        if 'author' in e.persons or 'editor' in e.persons:
            return toplevel[
                self.format_author_or_editor_and_date(e),
                self.format_title(e, 'title'),
                sentence(sep=' ')[
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional["(", pages, ")"]
                ],
                sentence()[
                    optional_field('publisher'),
                ],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                self.format_title(e, 'title'),
                self.format_date(e),
                sentence(sep=' ')[
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional["(", pages, ")"]
                ],
                sentence()[
                    optional_field('publisher'),
                ],
                sentence[optional_field('note')],
            ]

    def get_inproceedings_template(self, e):
        if 'author' in e.persons:
            return toplevel[
                self.format_author_or_editor_and_date(e),
                self.format_title(e, 'title'),
                sentence(sep=' ')[
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional["(", pages, ")"]
                ],
                sentence[optional_field('publisher')],
                sentence[optional_field('note')],
                self.format_web_refs(e)
            ]
        else:
            return toplevel[
                self.format_title(e, 'title'),
                self.format_date(e),
                sentence(sep=' ')[
                    self.format_btitle(e, 'booktitle', as_sentence=False),
                    optional["(", pages, ")"]
                ],
                sentence[optional_field('publisher')],
                sentence[optional_field('note')],
                self.format_web_refs(e)
            ]
    def get_manual_template(self, e):
        if 'author' in e.persons:
            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                self.format_btitle(e, 'title'),
                sentence[optional_field('address')],
                sentence[optional_field('note')],
                self.format_web_refs(e)
            ]
        else:
            return toplevel[
                self.format_btitle(e, 'title'),
                self.format_date(e),
                sentence[optional_field('address')],
                sentence[optional_field('note')],
                self.format_web_refs(e)
            ]

    def get_mastersthesis_template(self, e):
        return toplevel[
            sentence(sep=' ')[
                self.format_names('author'),
                self.format_date(e),
            ],
            sentence(sep=' ')[
                self.format_btitle(e, 'title', as_sentence=False),
                "(Master's thesis)"
            ],
            sentence[
                field('school'),
                optional_field('address'),
            ],
            sentence[optional_field('note')],
        ]

    def get_misc_template(self, e):
        if 'author' in e.persons:
            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                optional[self.format_btitle(e, 'title')],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                optional[self.format_btitle(e, 'title')],
                self.format_date(e),
                sentence[optional_field('note')],
            ]

    def get_phdthesis_template(self, e):
        return toplevel[
            sentence(sep=' ')[
                self.format_names('author'),
                self.format_date(e),
            ],
            sentence(sep=' ')[
                self.format_btitle(e, 'title', as_sentence=False),
                "(Doctoral dissertation)"
            ],
            sentence[
                field('school'),
                optional_field('address'),
            ],
            sentence[optional_field('note')],
        ]

    def get_proceedings_template(self, e):
        if 'editor' in e.persons:
            return toplevel[
                self.format_editor(e),
                self.format_date(e),
                self.format_btitle(e, 'title'),
                sentence[optional_field('publisher')],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                self.format_btitle(e, 'title'),
                self.format_date(e),
                sentence[optional_field('publisher')],
                sentence[optional_field('note')],
            ]

    def get_techreport_template(self, e):
        if 'author' in e.persons:
            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                self.format_btitle(e, 'title'),
                sentence[field('institution')],
                sentence[optional_field('note')],
            ]
        else:
            return toplevel[
                self.format_btitle(e, 'title'),
                self.format_date(e),
                sentence[field('institution')],
                sentence[optional_field('note')],
            ]

    def get_unpublished_template(self, e):
        if 'author' in e.persons:
            return toplevel[
                self.format_names('author'),
                self.format_date(e),
                self.format_btitle(e, 'title'),
                sentence[field('note')],
            ]
        else:
            return toplevel[
                self.format_btitle(e, 'title'),
                self.format_date(e),
                sentence[field('note')],
            ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['formatting', 'labels', 'names']

package_data = \
{'': ['*']}

install_requires = \
['pybtex>=0.24.0,<0.25.0']

entry_points = \
{'pybtex.style.formatting': ['apa7 = formatting.apa:APAStyle'],
 'pybtex.style.labels': ['apa7 = labels.apa:LabelStyle'],
 'pybtex.style.names': ['firstlast = names.firstlast:NameStyle']}

setup_kwargs = {
    'name': 'pybtex-apa7-style',
    'version': '0.1.1',
    'description': 'Provides APA7 style for Pybtex',
    'long_description': '# APA7 Style for Pybtex\n\n*This is a fork of [naeka\'s pybtex-apa-style](https://github.com/naeka/pybtex-apa-style)*.\n\n[Pybtex](https://pybtex.org/) provides Python support for interacting with bibtex-formatted\nbibliography information. Style plugins are required to format a bibliography in a given \nstyle, similar to the role that `csl` files play for LaTeX.\n\nThis plugin provides APA7 style.\n\n## Installation\n\n```\n$ pip install pybtex pybtex-apa7-style\n```\n\n## Usage\n\nPybtex uses [Python\'s plugin system](https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/).\nTo use APA7, load it as a plugin.\n\n```\nfrom pybtex.plugin import find_plugin\nfrom pybtex.database import parse_file\nAPA = find_plugin(\'pybtex.style.formatting\', \'apa7\')()\nHTML = find_plugin(\'pybtex.backends\', \'html\')()\n\ndef bib_to_apa7_html(bibfile):\n    bibliography = parse_file(bibfile, \'bibtex\')\n    formatted_bib = APA.format_bibliography(bibliography)\n    return "<br>".join(entry.text.render(HTML) for entry in formatted_bib)\n```\n',
    'author': 'Chris Proctor',
    'author_email': 'github.com@accounts.chrisproctor.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cproctor/pybtex-apa7-style',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
